import os
import tempfile
import xlrd
from urllib.request import urlretrieve
import pandas as pd
import geopandas as gpd


def get_column_table(silent=True) -> pd.DataFrame:
    """シェープファイルの列名の対応表を取得する。

    `シェープファイルの列名の対応表xls <http://nlftp.mlit.go.jp/ksj/gml/shape_property_table.xls>`__ を開き、ブック内の全シートを結合して単一のデータフレームにして返す。
    
    Returns
    -------
    pandas.DataFrame
        xlsファイル

    Example
    -------
    >>> import ksj
    >>> columns = ksj.get_column_table()
    >>> columns.head()
         識別子       データ項目                         タグ名      対応番号        属性名
    0  A02-a  指定地域3次メッシュ  DesignatedAreaTertiaryMesh  A02a_001  3次メッシュコード  NaN
    1  A02-a  指定地域3次メッシュ  DesignatedAreaTertiaryMesh  A02a_002   3次メッシュ面積  NaN
    2  A02-a  指定地域3次メッシュ  DesignatedAreaTertiaryMesh  A02a_003    農業地域の面積  NaN
    3  A02-a  指定地域3次メッシュ  DesignatedAreaTertiaryMesh  A02a_004    森林地域の面積  NaN
    4  A02-a  指定地域3次メッシュ  DesignatedAreaTertiaryMesh  A02a_005   都市計画区域面積  NaN
    """
    book = _get_shape_property_table(silent=silent)
    sheets = []
    for sheet_name in book.sheet_names():
        sheet = _get_sheet(book=book, sheet_name=sheet_name)
        sheets.append(sheet)
    return pd.concat(sheets, axis=0, sort=False)


def _get_shape_property_table(silent=True) -> xlrd.book.Book:
    """列名の対応表xlsファイルを一時フォルダにダウンロードし、開く。
    xlsファイルは開いてすぐ（メソッドの処理が終わるとともに）消去される
    """
    url = "http://nlftp.mlit.go.jp/ksj/gml/shape_property_table.xls"
    file_name = os.path.basename(url)
    temp_dir = tempfile.TemporaryDirectory()
    file_path = os.path.join(temp_dir.name, file_name)
    urlretrieve(url=url, filename=file_path)
    if not silent:
        print(f"shape_property_table is downloaded at {file_path}")
    book = xlrd.open_workbook(file_path)
    temp_dir.cleanup()
    return book


def _get_sheet(book: xlrd.book.Book, sheet_name=False) -> pd.DataFrame:
    """指定したsheet nameのシートをpandasで開く"""
    sheet = book.sheet_by_name(sheet_name)
    table = [sheet.row_values(row) for row in range(sheet.nrows)]
    index = _get_content_start_row_index(table)
    return pd.DataFrame(table[index+1:], columns=table[index])


def _get_content_start_row_index(table: list) -> int:
    """表頭っぽい行にきたらbreakして行番号を取得する"""
    for i in range(len(table)):
        if ("属性名" in table[i]) or ("データ項目" in table[i]):
            break
    return i


def translate(data: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """年度によって意味が変わらない列についてのみ列名を変換する。

    Parameters
    ----------
    data : geopandas.GeoDataFrame
        国土数値情報ダウンロードサービスから取得されたシェープファイル

    Returns
    -------
    geopandas.GeoDataFrame
        列名を日本語に変換したシェープファイル

    Example
    -------
    >>> import ksj
    >>> urls = ksj.get_url(identifier='N03', pref_code='13', fiscal_year=2019)
    >>> url = urls['zipFileUrl'][0]
    >>> shape_gdf = ksj.read_shp(url)
    >>> shape_gdf.head()
      N03_001 N03_002 N03_003 N03_004 N03_007                                           geometry
    0     東京都    None    None    千代田区   13101  POLYGON ((139.77287 35.70370, 139.77279 35.703...
    1     東京都    None    None     中央区   13102  POLYGON ((139.78341 35.69645, 139.78459 35.696...
    2     東京都    None    None      港区   13103  POLYGON ((139.77129 35.62841, 139.77128 35.628...
    3     東京都    None    None      港区   13103  POLYGON ((139.76689 35.62774, 139.76718 35.627...
    4     東京都    None    None      港区   13103  POLYGON ((139.77022 35.63199, 139.77046 35.631...
    >>> shape_gdf = ksj.translate(shape_gdf)
    >>> shape_gdf.head()
      都道府県名   支庁名 郡政令都市 市区町村名 行政区域コード                                           geometry
    0   東京都  None  None  千代田区   13101  POLYGON ((139.77287 35.70370, 139.77279 35.703...
    1   東京都  None  None   中央区   13102  POLYGON ((139.78341 35.69645, 139.78459 35.696...
    2   東京都  None  None    港区   13103  POLYGON ((139.77129 35.62841, 139.77128 35.628...
    3   東京都  None  None    港区   13103  POLYGON ((139.76689 35.62774, 139.76718 35.627...
    4   東京都  None  None    港区   13103  POLYGON ((139.77022 35.63199, 139.77046 35.631...
    """
    # 列名の対応表をダウンロードして結合して開く
    book = get_column_table()
    # TODO: 2713列中2470列をこれで対応できるが、残り243列（年度によって意味が変わる列コード）は今後対応する
    unique_cols = book[["対応番号", "属性名"]].drop_duplicates()
    codes = list(unique_cols["対応番号"])
    names = list(unique_cols["属性名"])
    rename_dict = dict(zip(codes, names))
    # rename
    return data.rename(columns=rename_dict)



import os
import tempfile
import zipfile
from urllib.request import urlretrieve
import geopandas as gpd
import re


def read_shp(url, return_type='auto'):
    """指定したURLのシェープファイルを取得して読み込む。
    
    指定したURLのzipファイルを一時フォルダにダウンロードし、解凍してGeoDataFrameで開く。
    稀に一つのzipファイルに複数のシェープファイルが含まれるものもあるが、その場合はGeoDataFrameのリストで返す。

    Parameters
    ----------
    url : str
        国土数値情報ダウンロードサービスが提供するzipファイルのURL
    return_type : str
        返り値の型を指定します。（default = "auto"）
        "auto"の場合、zipファイル内のシェープファイルが1つならGeoDataFrame、複数ならGeoDataFrameのlistで返します。
        "list"の場合、常にGeoDataFrameのlistを返します。

    Returns
    -------
    geopandas.GeoDataFrame or list
        読み込んだシェープファイルのデータ。
        一つのzipファイルに複数のシェープファイルがある場合、GeoDataFrameをlistに入れて返します。
    
    Example
    -------
    >>> import ksj
    >>> urls = ksj.get_url(identifier='N03', pref_code='13', fiscal_year=2019)
    >>> url = urls['zipFileUrl'][0]
    >>> shape_gdf = ksj.read_shp(url)
    >>> shape_gdf.head()
      N03_001 N03_002 N03_003 N03_004 N03_007                                           geometry
    0     東京都    None    None    千代田区   13101  POLYGON ((139.77287 35.70370, 139.77279 35.703...
    1     東京都    None    None     中央区   13102  POLYGON ((139.78341 35.69645, 139.78459 35.696...
    2     東京都    None    None      港区   13103  POLYGON ((139.77129 35.62841, 139.77128 35.628...
    3     東京都    None    None      港区   13103  POLYGON ((139.76689 35.62774, 139.76718 35.627...
    4     東京都    None    None      港区   13103  POLYGON ((139.77022 35.63199, 139.77046 35.631...
    """
    if (return_type != "auto") and (return_type != "list"):
        raise NameError("'return_type' must be 'auto' or 'list'")
    # download to temp dir
    zip_file_name = os.path.basename(url)
    temp_dir = tempfile.TemporaryDirectory()
    save_dir = temp_dir.name
    save_path = os.path.join(save_dir, zip_file_name)
    saved_path, response = urlretrieve(url=url, filename=save_path)
    # unzip
    name_without_extension = os.path.splitext(zip_file_name)[0]
    extract_dir = os.path.join(save_dir, name_without_extension)
    if not os.path.exists(extract_dir):
        os.mkdir(extract_dir)
    with zipfile.ZipFile(save_path) as existing_zip:
        # 文字化け回避のため変換処理を逐次挟む
        for info in existing_zip.infolist():
            info.filename = info.filename.encode('cp437').decode('cp932')
            existing_zip.extract(info, path=extract_dir)
    # load
    file_paths = _get_files(extract_dir)
    shapefiles = [f for f in file_paths if f.endswith(".shp")]
    if len(shapefiles) == 0:
        OSError("shapefile not found")
    else:
        shape_files = [_read_geofile(shp) for shp in shapefiles]
        if (len(shape_files) == 1) and (return_type != "list"):
            shape_files = shape_files[0]
    temp_dir.cleanup()
    return shape_files


def get_shp(url: str, path: str) -> None:
    """指定したURLのzipファイルを指定フォルダにダウンロードする。

    Parameters
    ----------
    url : str
        国土数値情報ダウンロードサービスが提供するzipファイルのURL
    path : str
        保存先のパス
    
    Example
    -------
    >>> import os
    >>> import ksj
    >>> url = "http://nlftp.mlit.go.jp/ksj/gml/data/N03/N03-2019/N03-190101_12_GML.zip"
    >>> save_dir = "/tmp/testing"
    >>> if not os.path.exists(save_dir):
    ...     os.mkdir(save_dir)
    ...
    >>> ksj.get_shp(url, path=save_dir)
    >>> os.listdir(save_dir)
    ['N03-190101_12_GML.zip']
    """
    try:
        urlretrieve(url=url, filename=path)
    except IsADirectoryError:
        zip_file_name = os.path.basename(url)
        urlretrieve(url=url, filename=os.path.join(path, zip_file_name))


def _read_geofile(file_path: str, verbose=1) -> gpd.GeoDataFrame:
    """
    geopandasで開ける地理データを読み込む
    既知の破損により読み込めないshpファイルは、同名のgeojsonがある場合そちらを読み込む
    """
    try:
        if verbose >= 2:
            print(f"reading a shapefile from {file_path}")
        gdf = gpd.read_file(file_path, encoding="CP932")
    except AttributeError:
        # A16-15_00_DID.shpはAttributeErrorで開けないが、同名のgeojsonは開ける
        # shpファイルと同名のgeojsonが同梱されているzipが124個ある
        if verbose >= 1:
            print(f"cannot read {file_path}")
        same_name_geojson = file_path.replace(".shp", ".geojson")
        if os.path.exists(same_name_geojson):
            if verbose >= 1:
                print(f"{same_name_geojson} found, trying to read it...")
            gdf = gpd.read_file(same_name_geojson, encoding="CP932")
            if verbose >= 1:
                print(f"finished.")
        else:
            gdf = None
    return gdf


def _get_files(path: str) -> list:
    """get file paths recursively"""
    file_names = []
    for (root, _, files) in os.walk(path):
        for f in files:
            file_names.append(os.path.join(root, f))
    return file_names



import pandas as pd
from lxml import etree
import xmljson
from urllib.request import Request, urlopen
from urllib.parse import urlencode
from urllib.error import HTTPError, URLError


def get_summary() -> pd.DataFrame:
    """国土数値情報の概要情報（取得できるデータ一覧）を取得する

    Returns
    -------
    pandas.DataFrame
        国土数値情報APIで取得できるデータの一覧
    
    Example
    -------
    >>> import ksj
    >>> ksj_summary = ksj.get_summary()
    >>> ksj_summary.head()
      identifier      title    field1 field2 areaType
    0        A03  三大都市圏計画区域      政策区域   大都市圏        2
    1        A09       都市地域  国土（水・土地）   土地利用        3
    2        A10     自然公園地域        地域   保護保全        3
    3        A11     自然保全地域        地域   保護保全        3
    4        A12       農業地域  国土（水・土地）   土地利用        3
    """
    # const params（2020年現在この値以外を受け付けない）
    app_id = "ksjapibeta1"
    lang = "J"
    data_format = 1
    # set url and params
    api_url = "http://nlftp.mlit.go.jp/ksj/api/1.0b/index.php/app/getKSJSummary.xml"
    # NOTE: 仕様書では'dataFormat'なのだが、実際は'dataformat'に送らないとエラーになる
    params = {"appId": app_id, "lang": lang, "dataformat": data_format}
    # request
    response, error = _get_request(api_url, params=params)
    if error is not None:
        print(error)
        return None
    root = etree.fromstring(response)
    # convert xml to dict
    data_dict = xmljson.yahoo.data(root)
    # error
    error_msg = data_dict["KSJ_SUMMARY_INF"]["RESULT"]["ERROR_MSG"]
    if error_msg != "正常に終了しました。":
        print(error_msg)
    else:
        # convert dict to pd.DataFrame
        data_df = pd.json_normalize(
            data_dict["KSJ_SUMMARY_INF"]["KSJ_SUMMARY"]["item"])
    return data_df


def get_url(identifier: str, pref_code=None, mesh_code=None,
            metro_area=None, fiscal_year=None) -> pd.DataFrame:
    """国土数値情報のURL情報を取得する

    Parameters
    ----------
    identifier : str
        ファイルの識別子（例：「地価公示」なら'L01'）
    pref_code : str, int
        都道府県コード（例：13, '11,12', '11-14'）
    mesh_code : str, int
        メッシュコード。areaType=4の場合のみ有効。（例：5340）
    metro_area : str, int
        都市圏コード。areaType=2の場合のみ有効。（例：100）
    fiscal_year : str, int
        年度。（例：2017, '2015,2019', '2014-2016'）

    Returns
    -------
    pandas.DataFrame
        データの情報とzipファイルのURLが入ったデータフレーム
    
    Example
    -------
    >>> import ksj
    >>> urls = ksj.get_url(identifier="N03", pref_code='11-14', fiscal_year=2019)
    >>> urls.head()
      identifier title field  year areaType areaCode datum                                         zipFileUrl zipFileSize
    0        N03  行政区域  政策区域  2019        3       11     1  http://nlftp.mlit.go.jp/ksj/gml/data/N03/N03-2...      3.54MB
    1        N03  行政区域  政策区域  2019        3       12     1  http://nlftp.mlit.go.jp/ksj/gml/data/N03/N03-2...      6.17MB
    2        N03  行政区域  政策区域  2019        3       13     1  http://nlftp.mlit.go.jp/ksj/gml/data/N03/N03-2...     12.20MB
    3        N03  行政区域  政策区域  2019        3       14     1  http://nlftp.mlit.go.jp/ksj/gml/data/N03/N03-2...      5.22MB
    """
    # const params
    app_id = "ksjapibeta1"
    lang = "J"
    data_format = 1
    # set url and params
    api_url = "http://nlftp.mlit.go.jp/ksj/api/1.0b/index.php/app/getKSJURL.xml"
    params = {"appId": app_id, "lang": lang,
              "dataformat": data_format, "identifier": identifier}
    if pref_code:
        params["prefCode"] = pref_code
    if mesh_code:
        params["meshCode"] = mesh_code
    if metro_area:
        params["metroArea"] = metro_area
    if fiscal_year:
        params["fiscalyear"] = fiscal_year
    # request
    response, error = _get_request(api_url, params=params)
    if error is not None:
        print(error)
        return None
    # convert to xml
    root = etree.fromstring(response)
    # convert xml to dict
    data_dict = xmljson.yahoo.data(root)

    error_msg = data_dict["KSJ_URL_INF"]["RESULT"]["ERROR_MSG"]
    if error_msg != "正常に終了しました。":
        print(error_msg)
        return None
    else:
        # convert dict to pd.DataFrame
        data_df = pd.json_normalize(
            data_dict["KSJ_URL_INF"]["KSJ_URL"]["item"])
        return data_df


def _get_request(url: str, params: dict) -> list:
    """リクエストパラメータ付きでGETリクエストを送る"""
    url = url + "?" + urlencode(params)
    response, error = None, None
    try:
        req = Request(url)
        with urlopen(req) as res:
            response = res.read()
    except HTTPError as e:
        error = e
    except URLError as e:
        error = e
    return response, error
